@@ -304,7 +304,7 @@ int gnutls_bye(gnutls_session_t session, gnutls_close_request_t how)
 			gnutls_assert();
 			return ret;
 		}
-
+		/* fallthrough */
 	case STATE61:
 		ret =
 		    gnutls_alert_send(session, GNUTLS_AL_WARNING,
@@ -321,7 +321,7 @@ int gnutls_bye(gnutls_session_t session, gnutls_close_request_t how)
 			do {
 				ret =
 				    _gnutls_recv_int(session, GNUTLS_ALERT,
-						     -1, NULL, 0, NULL,
+						     -1, NULL, NULL, 0, NULL,
 						     session->internals.
 						     record_timeout_ms);
 			}
@@ -635,6 +635,34 @@ check_buffers(gnutls_session_t session, content_type_t type,
 	return 0;
 }
 
+/* Checks and retrieves any pending data in the application data record buffers.
+ */
+static int
+check_packet_buffers(gnutls_session_t session, content_type_t type,
+		     gnutls_packet_t *packet)
+{
+	if (_gnutls_record_buffer_get_size(session) > 0) {
+		int ret;
+		ret =
+		    _gnutls_record_buffer_get_packet(type, session,
+					      	     packet);
+		if (ret < 0) {
+			if (IS_DTLS(session)) {
+				if (ret == GNUTLS_E_UNEXPECTED_PACKET) {
+					ret = GNUTLS_E_AGAIN;
+				}
+			}
+			gnutls_assert();
+			return ret;
+		}
+
+		return ret;
+	}
+
+	*packet = NULL;
+	return 0;
+}
+
 
 
 /* Here we check if the advertized version is the one we
@@ -784,10 +812,10 @@ record_add_to_buffers(gnutls_session_t session,
 			break;
 
 		case GNUTLS_CHANGE_CIPHER_SPEC:
-			if (!(IS_DTLS(session)))
-				return
-				    gnutls_assert_val
-				    (GNUTLS_E_UNEXPECTED_PACKET);
+			if (!(IS_DTLS(session))) {
+				ret = gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET);
+				goto cleanup;
+			}
 
 			_gnutls_record_buffer_put(session, recv->type, seq,
 						  bufel);
@@ -818,16 +846,8 @@ record_add_to_buffers(gnutls_session_t session,
 			    || (htype == GNUTLS_HANDSHAKE_CLIENT_HELLO
 				&& type == GNUTLS_HANDSHAKE)) {
 				/* even if data is unexpected put it into the buffer */
-				if ((ret =
-				     _gnutls_record_buffer_put(session,
-							       recv->type,
-							       seq,
-							       bufel)) <
-				    0) {
-					gnutls_assert();
-					goto cleanup;
-				}
-
+				_gnutls_record_buffer_put(session, recv->type,
+							  seq, bufel);
 				return
 				    gnutls_assert_val
 				    (GNUTLS_E_GOT_APPLICATION_DATA);
@@ -891,14 +911,9 @@ record_add_to_buffers(gnutls_session_t session,
 			    && bufel->htype ==
 			    GNUTLS_HANDSHAKE_CLIENT_HELLO) {
 				gnutls_assert();
-				ret =
-				    _gnutls_record_buffer_put(session,
+				_gnutls_record_buffer_put(session,
 							      recv->type,
 							      seq, bufel);
-				if (ret < 0) {
-					gnutls_assert();
-					goto cleanup;
-				}
 				return GNUTLS_E_REHANDSHAKE;
 			}
 
@@ -1296,9 +1311,9 @@ _gnutls_recv_in_buffers(gnutls_session_t session, content_type_t type,
 		goto begin;
 	}
 
-	if (record.v2)
+	if (record.v2) {
 		decrypted->htype = GNUTLS_HANDSHAKE_CLIENT_HELLO_V2;
-	else {
+	} else {
 		uint8_t *p = _mbuffer_get_udata_ptr(decrypted);
 		decrypted->htype = p[0];
 	}
@@ -1307,7 +1322,7 @@ _gnutls_recv_in_buffers(gnutls_session_t session, content_type_t type,
 	    record_add_to_buffers(session, &record, type, htype,
 				  packet_sequence, decrypted);
 
-	/* bufel is now either deinitialized or buffered somewhere else */
+	/* decrypted is now either deinitialized or buffered somewhere else */
 
 	if (ret < 0)
 		return gnutls_assert_val(ret);
@@ -1373,12 +1388,13 @@ _gnutls_recv_in_buffers(gnutls_session_t session, content_type_t type,
 ssize_t
 _gnutls_recv_int(gnutls_session_t session, content_type_t type,
 		 gnutls_handshake_description_t htype,
+		 gnutls_packet_t *packet,
 		 uint8_t * data, size_t data_size, void *seq,
 		 unsigned int ms)
 {
 	int ret;
 
-	if ((type != GNUTLS_ALERT && type != GNUTLS_HEARTBEAT)
+	if (packet == NULL && (type != GNUTLS_ALERT && type != GNUTLS_HEARTBEAT)
 	    && (data_size == 0 || data == NULL))
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
@@ -1402,23 +1418,110 @@ _gnutls_recv_int(gnutls_session_t session, content_type_t type,
 	case RECV_STATE_0:
 
 		_dtls_async_timer_check(session);
-		/* If we have enough data in the cache do not bother receiving
-		 * a new packet. (in order to flush the cache)
-		 */
-		ret = check_buffers(session, type, data, data_size, seq);
-		if (ret != 0)
-			return ret;
 
-		ret = _gnutls_recv_in_buffers(session, type, htype, ms);
-		if (ret < 0 && ret != GNUTLS_E_SESSION_EOF)
-			return gnutls_assert_val(ret);
+		if (packet == NULL) {
+			/* If we have enough data in the cache do not bother receiving
+			 * a new packet. (in order to flush the cache)
+			 */
+			ret = check_buffers(session, type, data, data_size, seq);
+			if (ret != 0)
+				return ret;
 
-		return check_buffers(session, type, data, data_size, seq);
+			ret = _gnutls_recv_in_buffers(session, type, htype, ms);
+			if (ret < 0 && ret != GNUTLS_E_SESSION_EOF)
+				return gnutls_assert_val(ret);
+
+			return check_buffers(session, type, data, data_size, seq);
+		} else {
+			ret = check_packet_buffers(session, type, packet);
+			if (ret != 0)
+				return ret;
+
+			ret = _gnutls_recv_in_buffers(session, type, -1, ms);
+			if (ret < 0 && ret != GNUTLS_E_SESSION_EOF)
+				return gnutls_assert_val(ret);
+
+			return check_packet_buffers(session, type, packet);
+		}
 	default:
 		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
 	}
 }
 
+/**
+ * gnutls_packet_get:
+ * @packet: is a #gnutls_packet_t structure.
+ * @data: will contain the data present in the @packet structure (may be %NULL)
+ * @sequence: the 8-bytes of the packet sequence number (may be %NULL)
+ *
+ * This function returns the data and sequence number associated with
+ * the received packet.
+ *
+ * Since: 3.3.5
+ **/
+
+void gnutls_packet_get(gnutls_packet_t packet, gnutls_datum_t *data, unsigned char *sequence)
+{
+	if (unlikely(packet == NULL)) {
+		gnutls_assert();
+		if (data) {
+			data->data = NULL;
+			data->size = 0;
+		}
+	}
+
+	if (sequence) {
+		memcpy(sequence, packet->record_sequence.i, 8);
+	}
+
+	if (data) {
+		data->size = packet->msg.size - packet->mark;
+		data->data = packet->msg.data + packet->mark;
+	}
+}
+
+/**
+ * gnutls_packet_deinit:
+ * @packet: is a pointer to a #gnutls_packet_st structure.
+ *
+ * This function will deinitialize all data associated with
+ * the received packet.
+ *
+ * Since: 3.3.5
+ **/
+void gnutls_packet_deinit(gnutls_packet_t packet)
+{
+	gnutls_free(packet);
+}
+
+/**
+ * gnutls_record_recv_packet:
+ * @session: is a #gnutls_session_t structure.
+ * @packet: the structure that will hold the packet data
+ *
+ * This is a lower-level function thatn gnutls_record_recv() and allows
+ * to directly receive the whole decrypted packet. That avoids a
+ * memory copy, and is mostly applicable to applications seeking high
+ * performance.
+ *
+ * The received packet is accessed using gnutls_packet_get() and 
+ * must be deinitialized using gnutls_packet_deinit(). The returned
+ * packet will be %NULL if the return value is zero (EOF).
+ *
+ * Returns: The number of bytes received and zero on EOF (for stream
+ * connections).  A negative error code is returned in case of an error.  
+ *
+ * Since: 3.3.5
+ **/
+ssize_t
+gnutls_record_recv_packet(gnutls_session_t session, 
+		   	  gnutls_packet_t *packet)
+{
+	return _gnutls_recv_int(session, GNUTLS_APPLICATION_DATA, -1, packet,
+				NULL, 0, NULL,
+				session->internals.record_timeout_ms);
+}
+
 /**
  * gnutls_record_send:
  * @session: is a #gnutls_session_t structure.
@@ -1599,8 +1702,8 @@ int gnutls_record_uncork(gnutls_session_t session, unsigned int flags)
 ssize_t
 gnutls_record_recv(gnutls_session_t session, void *data, size_t data_size)
 {
-	return _gnutls_recv_int(session, GNUTLS_APPLICATION_DATA, -1, data,
-				data_size, NULL,
+	return _gnutls_recv_int(session, GNUTLS_APPLICATION_DATA, -1, NULL,
+				data, data_size, NULL,
 				session->internals.record_timeout_ms);
 }
 
@@ -1628,8 +1731,8 @@ ssize_t
 gnutls_record_recv_seq(gnutls_session_t session, void *data,
 		       size_t data_size, unsigned char *seq)
 {
-	return _gnutls_recv_int(session, GNUTLS_APPLICATION_DATA, -1, data,
-				data_size, seq,
+	return _gnutls_recv_int(session, GNUTLS_APPLICATION_DATA, -1, NULL,
+				data, data_size, seq,
 				session->internals.record_timeout_ms);
 }
 
