@@ -41,13 +41,28 @@
 #include "certtool-common.h"
 #include "certtool-args.h"
 #include "certtool-cfg.h"
+#include <minmax.h>
 
 /* Gnulib portability files. */
 #include <read-file.h>
 
-unsigned char buffer[64 * 1024];
-const int buffer_size = sizeof(buffer);
+unsigned char *lbuffer = NULL;
+int lbuffer_size = 0;
 
+void fix_lbuffer(unsigned size)
+{
+	if (lbuffer_size == 0 || lbuffer == NULL) {
+		if (size == 0)
+			lbuffer_size = 64*1024;
+		else
+			lbuffer_size = MAX(64*1024,size);
+		lbuffer = malloc(lbuffer_size);
+		if (lbuffer == NULL) {
+			fprintf(stderr, "memory error");
+			exit(1);
+		}
+	}
+}
 
 FILE *safe_open_rw(const char *file, int privkey_op)
 {
@@ -119,7 +134,7 @@ const char *get_password(common_info_st * cinfo, unsigned int *flags,
 			p = get_pass();
 	}
 
-	if (p[0] == 0 && flags && !cinfo->empty_password)
+	if ((p == NULL || p[0] == 0) && flags && !cinfo->empty_password)
 		*flags |= GNUTLS_PKCS_PLAIN;
 
 	return p;
@@ -351,6 +366,8 @@ gnutls_x509_crt_t *load_cert_list(int mand, size_t * crt_size,
 	size_t size;
 	int ptr_size;
 
+	fix_lbuffer(0);
+
 	*crt_size = 0;
 	if (info->verbose)
 		fprintf(stderr, "Loading certificate list...\n");
@@ -365,16 +382,16 @@ gnutls_x509_crt_t *load_cert_list(int mand, size_t * crt_size,
 
 	fd = fopen(info->cert, "r");
 	if (fd == NULL) {
-		fprintf(stderr, "%s\n", info->cert);
+		fprintf(stderr, "Could not open %s\n", info->cert);
 		exit(1);
 	}
 
-	size = fread(buffer, 1, sizeof(buffer) - 1, fd);
-	buffer[size] = 0;
+	size = fread(lbuffer, 1, lbuffer_size - 1, fd);
+	lbuffer[size] = 0;
 
 	fclose(fd);
 
-	ptr = (void *) buffer;
+	ptr = (void *) lbuffer;
 	ptr_size = size;
 
 	for (i = 0; i < MAX_CERTS; i++) {
@@ -407,7 +424,7 @@ gnutls_x509_crt_t *load_cert_list(int mand, size_t * crt_size,
 		ptr_size = size;
 		ptr_size -=
 		    (unsigned int) ((unsigned char *) ptr -
-				    (unsigned char *) buffer);
+				    (unsigned char *) lbuffer);
 
 		if (ptr_size < 0)
 			break;
@@ -498,13 +515,17 @@ gnutls_privkey_t load_ca_private_key(common_info_st * info)
 
 /* Loads the CA's certificate
  */
-gnutls_x509_crt_t load_ca_cert(common_info_st * info)
+gnutls_x509_crt_t load_ca_cert(unsigned mand, common_info_st * info)
 {
 	gnutls_x509_crt_t crt;
 	int ret;
 	gnutls_datum_t dat;
 	size_t size;
 
+	if (mand == 0 && info->ca == NULL) {
+		return NULL;
+	}
+
 	if (info->ca == NULL) {
 		fprintf(stderr, "missing --load-ca-certificate\n");
 		exit(1);
@@ -520,7 +541,7 @@ gnutls_x509_crt_t load_ca_cert(common_info_st * info)
 	dat.size = size;
 
 	if (!dat.data) {
-		fprintf(stderr, "reading --load-ca-certificate: %s\n",
+		fprintf(stderr, "error reading --load-ca-certificate: %s\n",
 			info->ca);
 		exit(1);
 	}
@@ -629,15 +650,17 @@ get_bits(gnutls_pk_algorithm_t key_type, int info_bits,
 		}
 		bits = info_bits;
 	} else {
-		if (info_sec_param) {
-			bits =
-			    gnutls_sec_param_to_pk_bits(key_type,
-							str_to_sec_param
-							(info_sec_param));
-		} else
-			bits =
-			    gnutls_sec_param_to_pk_bits(key_type,
-							GNUTLS_SEC_PARAM_NORMAL);
+		if (info_sec_param == 0) {
+			/* For ECDSA keys use 256 bits or better, as they are widely supported */
+			if (key_type == GNUTLS_PK_EC)
+				info_sec_param = "HIGH";
+			else
+				info_sec_param = "MEDIUM";
+		}
+		bits =
+		    gnutls_sec_param_to_pk_bits(key_type,
+						str_to_sec_param
+						(info_sec_param));
 	}
 
 	return bits;
@@ -649,8 +672,8 @@ gnutls_sec_param_t str_to_sec_param(const char *str)
 		return GNUTLS_SEC_PARAM_LOW;
 	} else if (strcasecmp(str, "legacy") == 0) {
 		return GNUTLS_SEC_PARAM_LEGACY;
-	} else if (strcasecmp(str, "normal") == 0) {
-		return GNUTLS_SEC_PARAM_NORMAL;
+	} else if (strcasecmp(str, "normal") == 0 || strcasecmp(str, "medium") == 0) {
+		return GNUTLS_SEC_PARAM_MEDIUM;
 	} else if (strcasecmp(str, "high") == 0) {
 		return GNUTLS_SEC_PARAM_HIGH;
 	} else if (strcasecmp(str, "ultra") == 0) {
@@ -798,6 +821,8 @@ void _pubkey_info(FILE * outfile,
 	int ret;
 	size_t size;
 
+	fix_lbuffer(0);
+
 	ret = gnutls_pubkey_print(pubkey, format, &data);
 	if (ret < 0) {
 		fprintf(stderr, "pubkey_print error: %s\n",
@@ -808,16 +833,16 @@ void _pubkey_info(FILE * outfile,
 	fprintf(outfile, "%s\n", data.data);
 	gnutls_free(data.data);
 
-	size = buffer_size;
+	size = lbuffer_size;
 	ret =
-	    gnutls_pubkey_export(pubkey, GNUTLS_X509_FMT_PEM, buffer,
+	    gnutls_pubkey_export(pubkey, GNUTLS_X509_FMT_PEM, lbuffer,
 				 &size);
 	if (ret < 0) {
 		fprintf(stderr, "export error: %s\n", gnutls_strerror(ret));
 		exit(1);
 	}
 
-	fprintf(outfile, "\n%s\n", buffer);
+	fprintf(outfile, "\n%s\n", lbuffer);
 }
 
 static void
@@ -853,6 +878,8 @@ void dh_info(FILE * infile, FILE * outfile, common_info_st * ci)
 	gnutls_datum_t p, g;
 	unsigned int q_bits = 0;
 
+	fix_lbuffer(0);
+
 	if (gnutls_dh_params_init(&dh_params) < 0) {
 		fprintf(stderr, "Error in dh parameter initialization\n");
 		exit(1);
@@ -881,18 +908,18 @@ void dh_info(FILE * infile, FILE * outfile, common_info_st * ci)
 		print_dh_info(outfile, &p, &g, q_bits, ci->cprint);
 
 	if (!ci->cprint) {	/* generate a PKCS#3 structure */
-		size_t len = buffer_size;
+		size_t len = lbuffer_size;
 
 		ret =
 		    gnutls_dh_params_export_pkcs3(dh_params,
 						  ci->outcert_format,
-						  buffer, &len);
+						  lbuffer, &len);
 
 		if (ret == 0) {
 			if (ci->outcert_format == GNUTLS_X509_FMT_PEM) {
-				fprintf(outfile, "\n%s", buffer);
+				fprintf(outfile, "\n%s", lbuffer);
 			} else {
-				fwrite(buffer, 1, len, outfile);
+				fwrite(lbuffer, 1, len, outfile);
 			}
 		} else {
 			fprintf(stderr, "Error: %s\n",
@@ -913,6 +940,8 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 	int bits = get_bits(GNUTLS_PK_DH, info->bits, info->sec_param, 1);
 	unsigned int q_bits = 0;
 
+	fix_lbuffer(0);
+
 	gnutls_dh_params_init(&dh_params);
 
 	if (how != 0) {
@@ -979,15 +1008,15 @@ int generate_prime(FILE * outfile, int how, common_info_st * info)
 	print_dh_info(outfile, &p, &g, q_bits, info->cprint);
 
 	if (!info->cprint) {	/* generate a PKCS#3 structure */
-		size_t len = buffer_size;
+		size_t len = lbuffer_size;
 
 		ret =
 		    gnutls_dh_params_export_pkcs3(dh_params,
 						  GNUTLS_X509_FMT_PEM,
-						  buffer, &len);
+						  lbuffer, &len);
 
 		if (ret == 0) {
-			fprintf(outfile, "\n%s", buffer);
+			fprintf(outfile, "\n%s", lbuffer);
 		} else {
 			fprintf(stderr, "Error: %s\n",
 				gnutls_strerror(ret));
